package de.evoal.core.arff.main;

import de.evoal.core.api.cdi.Application;
import de.evoal.core.api.cdi.BlackboardValue;
import de.evoal.core.api.cdi.MainClass;
import de.evoal.core.arff.cdi.ArffBlackboardEntries;
import lombok.extern.slf4j.Slf4j;
import weka.core.Attribute;
import weka.core.Instances;
import weka.core.converters.ConverterUtils;

import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;
import java.io.*;
import java.util.Enumeration;

@Slf4j
@ApplicationScoped
@Application(
        name = "extract-data-definition-from-arff",
        documentation = "Application to automatically create a data definition file (output) based on an arff file (input)."
)
public class ExtractDataDefinition implements MainClass {

    @Inject
    @BlackboardValue(ArffBlackboardEntries.ARFF_INPUT)
    private String arffFile;

    @Inject
    @BlackboardValue(ArffBlackboardEntries.OUTPUT_FILE)
    private File ddlFile;

    @Override
    public void run() {
        try {
            final ConverterUtils.DataSource source = new ConverterUtils.DataSource(arffFile);
            final Instances structure = source.getStructure();

            try(final OutputStream ddlStream = new FileOutputStream(ddlFile)) {
                try(final PrintStream ddlPrinter = new PrintStream(ddlStream)) {
                    extractDDL(ddlPrinter, structure);
                }
            }
        } catch(final Exception e) {
            log.error("Failed to extract ddl from {}. {} might be corrupt.", arffFile, ddlFile, e);
        }
    }

    public static void extractDDL(final PrintStream writer, final Instances structure) {
        final Enumeration<Attribute> attributes = structure.enumerateAttributes();
        writer.println("data:");

        while(attributes.hasMoreElements()) {
            final Attribute attr = attributes.nextElement();

            writer.print("    ");
            writer.print(toType(attr.type()));
            writer.print(" ");
            writer.print(toStorage(attr.type()));
            writer.print(" ");
            writer.print("data");
            writer.print(" ");
            writer.print("'");
            writer.print(attr.name());
            writer.print("'");
            writer.println(";");
        }
    }

    private static String toStorage(int type) {
        switch (type) {
            case Attribute.NOMINAL, Attribute.STRING -> {
                return "string";
            }

            case Attribute.NUMERIC -> {
                return "real";
            }

            default -> {
                log.error("Unsupported storage type: {}", type);
                return "???";
            }
        }

    }

    private static String toType(int type) {
        switch (type) {
            case Attribute.NOMINAL, Attribute.STRING -> {
                return "nominal";
            }
            case Attribute.NUMERIC -> {
                return "cardinal";
            }
            default -> {
                log.error("Unsupported type: {}", type);
                return "???";
            }
        }
    }
}
